# coding: utf-8
"""
Модуль функций подсчета показателей.
"""

from dateutil.relativedelta import relativedelta
import datetime

from django.db.models import Q

from kinder.core.declaration_status.models import DSS
from kinder.core.declaration.models import Declaration, DeclarationUnit
from kinder.core.direct.models import Direct
from kinder.core.dict.models import UnitKind
from kinder.core.direct.models import DirectStatusSewn as DiDD

from gisdo import settings as gs_settings

from gisdo.constants import (
    AGE_CATEGORIES_CALC,
    AGE_CATEGORIES_FULL,
    AGE_CATEGORIES_CUT,
    AGE_CATEGORIES_EIGHT,
    ALL)
from gisdo.utils import AgeDeltas

# статусы заявлений для запросов
APPROVED_STATUS_CODES = [DSS.ZAGS_CHECKING, DSS.DUL_CONFIRMATING,
                         DSS.PRIV_CONFIRMATING, DSS.MED_CONFIRMATING,
                         DSS.TUTOR_CONFIRMATING, DSS.WANT_CHANGE_DOU,
                         DSS.REGISTERED]
REFUSED_STATUS_CODES = [DSS.REFUSED]
EXCLUDE_STATUSES = [DSS.ACCEPTED, DSS.ARCHIVE]


class QueueCacheBuilder(object):
    """
    Сборка кэша очереди.
    """

    DECLARATION_FIELD_LIST = [
        'id', 'work_type__code', 'status__code',
        'desired_date', 'best_privilege__id',
        'children__health_need_id',
        'children__date_of_birth', 'mo', 'date']

    @staticmethod
    def _get_most_priority_unit(declaration):
        """
        Тупо и храбро выясняем самый приоритетное учреждение заявки
        """

        declaration_units = DeclarationUnit.objects.filter(
            declaration__id=declaration['id'],
            unit__kind__id=UnitKind.DOU).order_by('ord')
        return declaration_units[0] if declaration_units.exists() else None

    @staticmethod
    def _get_directed_declarations(declarations):
        """
        Заявки в статусе Направлен в ДОУ
        @ declarations QuerySet
        """

        return declarations.filter(
            status__code=DSS.DIRECTED).values(
                *QueueCacheBuilder.DECLARATION_FIELD_LIST)

    @staticmethod
    def _get_declaration_with_direction(declarations):
        """
        Заявки в других статусах, у которых есть привязанные направления
        в статусах "Заключение договора", "Подтверждено",
        "Направлен в ДОУ", "Предложено системой".
        @ declarations QuerySet
        """

        declarations = declarations.select_related(
            'children', 'best_privilege',
            'work_type', 'status')

        declarations = declarations.filter(
            direct__status__code__in=[
                DiDD.DOGOVOR, DiDD.CONFIRM,
                DiDD.REGISTER, DiDD.NEW]
        ).exclude(
            status__code=DSS.DIRECTED
        )

        declarations = declarations.exclude(
            status__code__in=[DSS.REGISTERED, DSS.PRIV_CONFIRMATING],
            direct__status__code=DiDD.NEW
        )

        return declarations.values(
            *QueueCacheBuilder.DECLARATION_FIELD_LIST)

    @staticmethod
    def _get_common_queue(declarations):

        declarations = (
            declarations.filter(
                status__code__in=[
                    DSS.REGISTERED, DSS.PRIV_CONFIRMATING]
            ).select_related(
                'children', 'best_privilege',
                'work_type', 'status'))
        direct = Direct.objects.exclude(
            status__code__in=[
                DiDD.ACCEPT, DiDD.REJECT,
                DiDD.NEED_CHANGE, DiDD.NEW, DiDD.REFUSED]).values(
                    'declaration_id')

        # Отсеиваем все заявления у ребенка имеющего временные зачисления
        if gs_settings.NO_TEMPORARY_PUPIL:
            declarations = declarations.exclude(
                children__pupil__temporary=True).distinct()

        return declarations.filter(~Q(id__in=direct)).values(
            *QueueCacheBuilder.DECLARATION_FIELD_LIST)

    @staticmethod
    def _get_declarations(unit):
        """
        Получаем заявки в очереди
        Очередь определена в задаче
        http://jira.bars-open.ru/browse/EDUKNDGFR-2
        """

        declarations = Declaration.objects.all()
        if unit.kind.id != UnitKind.REGION:
            declarations = declarations.filter(mo=unit)

        results = list(
            QueueCacheBuilder._get_directed_declarations(
                declarations))
        results.extend(
            QueueCacheBuilder._get_declaration_with_direction(
                declarations))
        results.extend(QueueCacheBuilder._get_common_queue(
            declarations))

        return results

    @classmethod
    def build_cache(cls, unit):

        queue_cache = {}
        age_categories = AGE_CATEGORIES_FULL.values()
        age_categories.extend(AGE_CATEGORIES_CUT.values())
        age_categories.extend(AGE_CATEGORIES_EIGHT.values())
        age_categories.append(ALL)

        declarations = QueueCacheBuilder._get_declarations(unit)
        for declaration in declarations:

            # TODO: Запрос в цикле. Если возникают
            # TODO: проблемы с производительностью, то переносим память.
            most_priority_unit = QueueCacheBuilder._get_most_priority_unit(
                declaration)

            if most_priority_unit:

                if most_priority_unit.unit.id not in queue_cache:
                    queue_cache[most_priority_unit.unit.id] = {
                        'CURRENT': {},
                        'ON_NEW_YEAR': {}
                    }

                date_of_birth = declaration['children__date_of_birth']

                for age_category in age_categories:

                    if (age_category not in queue_cache[
                            most_priority_unit.unit.id]['CURRENT'] or
                        age_category not in queue_cache[
                            most_priority_unit.unit.id]['ON_NEW_YEAR']):
                        queue_cache[
                            most_priority_unit.unit.id]['CURRENT'][
                                age_category] = []
                        queue_cache[
                            most_priority_unit.unit.id]['ON_NEW_YEAR'][
                                age_category] = []

                    birth_date_calculate = datetime.date.today()
                    down_border = birth_date_calculate - relativedelta(
                        years=AGE_CATEGORIES_CALC[age_category][1],
                        months=AGE_CATEGORIES_CALC[age_category][3])
                    up_border = birth_date_calculate - relativedelta(
                        years=AGE_CATEGORIES_CALC[age_category][0],
                        months=AGE_CATEGORIES_CALC[age_category][2])
                    if down_border < date_of_birth <= up_border:
                        queue_cache[
                            most_priority_unit.unit.id]['CURRENT'][
                                age_category].append(declaration)

                    birth_date_calculate = AgeDeltas.calculate_date()
                    down_border = birth_date_calculate - relativedelta(
                        years=AGE_CATEGORIES_CALC[age_category][1],
                        months=AGE_CATEGORIES_CALC[age_category][3])
                    up_border = birth_date_calculate - relativedelta(
                        years=AGE_CATEGORIES_CALC[age_category][0],
                        months=AGE_CATEGORIES_CALC[age_category][2])
                    if down_border < date_of_birth <= up_border:
                        queue_cache[
                            most_priority_unit.unit.id]['ON_NEW_YEAR'][
                                age_category].append(declaration)

        return queue_cache